include ("StationQuestHelpers");
include( "MathHelpers" );
include("ArtifactUtilities.lua");


-- ==========================================================
-- Globals
-- ==========================================================
g_Stations = {};

local BASE_WITHDRAWAL_TURN_THRESHOLD = GameDefines["STATION_WITHDRAWAL_TURN_THRESHOLD"];
local DEFAULT_STATION_MAX_LEVEL = GameDefines["STATION_DEFAULT_MAX_LEVEL"];
local CircuitChanceBase = 15 -- Chance for an Artifact after moving back and forth between Station and City
local CapitalTradeRouteCompleteBase = 30 -- How much Capital is gained after completing a Trade Route

local iResearchMod = GameInfo.GameSpeeds[Game.GetGameSpeedType()].ResearchPercent / 100
local CircuitChance = CircuitChanceBase / iResearchMod
local CapitalTradeRouteComplete = CapitalTradeRouteCompleteBase * iResearchMod

local TextEventTradeCompleteHeadline = "Station Trade Route Complete"
local TextEventTradeComplete = "The Station gained a Level (if it wasn't already on Level 5), you were granted an Artifact and gained " .. CapitalTradeRouteComplete .. " Diplomatic Capital."

-- Checks if the station script has overridden the process for leveling up. To return true,
-- the station must override BOTH the UpdateStationLevel and GetLevelString methods
local CHECK_DEFAULT_LEVELING_OVERRIDE = function(station)
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.UpdateStationLevel ~= nil and stationTable.GetLevelString ~= nil) then
		return true;
	else
		return false;
	end
end

local DEFAULT_UPDATE_LEVEL = function(station)
	local numTradeRoutesCompleted = station:GetNumTradeRoutesCompleted();
	local newLevel = numTradeRoutesCompleted + 1;
	local maxLevel = GetStationMaxLevel(station:GetID());
	if (newLevel <= maxLevel) then
		station:SetLevel(newLevel);
	end
end

local DEFAULT_GET_LEVEL_STRING = function(station, playerType)

	local stationInfo = GameInfo.Stations[station:GetType()];
	local stationTable = GetStationTable(station:GetType());
	local currentLevel = station:GetLevel();
	local maxLevel = GetStationMaxLevel(station:GetID());

	if (currentLevel >= maxLevel) then
		return Locale.ConvertTextKey("TXT_KEY_STATION_DEFAULT_MAX_LEVEL_MESSAGE", stationInfo.Description, currentLevel);
	else
		local nextLevel = currentLevel + 1;
		local currentLevelMessage = Locale.ConvertTextKey("TXT_KEY_STATION_DEFAULT_LEVEL_MESSAGE", stationInfo.Description, currentLevel, nextLevel);
		local nextLevelMessage = GetNextLevelString(station:GetID(), playerType);
		return currentLevelMessage .. "[NEWLINE][NEWLINE]" .. nextLevelMessage;
	end
end

-- ==========================================================

-- System Methods

-- ==========================================================
function ShouldDoStationSpawn()
	
	

	-- Must be past the minimum game turn threshold
	-- Link this to game speed build percent
	local baseStartTurn = (GameDefines.STATION_SPAWN_BASE_START_TURN * Game:GetBuildPercent()) / 100;
	if Game.GetGameTurn() < baseStartTurn then
		return false;
	end

	-- Must be at least one candidate plot
	if MINORPOWERS.GetNumStationCandidatePlots() <= 0 then
		return false;
	end

	-- Must be at least two available stations (to have the choice)
	if MINORPOWERS.GetNumAvailableStationTypes() < 2 then
		return false;
	end

	-- Check the ratio of active stations to players
	local numActiveStations = Game.GetNumActiveStations();
	local numActiveCivs = Game.CountMajorCivsActive();
	local ratio = ZeroSafeRatio(numActiveStations, numActiveCivs);
	if ratio >= GameDefines.STATION_TO_PLAYER_DEFAULT_MAX_RATIO then
		return false;
	end

	local turnsSinceLastSpawn = MINORPOWERS.GetTurnsSinceLastStationSpawn();
	local currentSpawnRate = MINORPOWERS.GetCurrentSpawnRate();

	local turnsOver = turnsSinceLastSpawn - currentSpawnRate;
	if turnsOver > 0 then

		-- Increase chance to spawn by 10% for each turn over the current rate (past the first), minimum 50%
		local chance : number = math.max(50, turnsOver * 10);
		local roll : number = Game.Rand(100, "Station Spawn chance roll");
		if (roll <= chance) then
			return true;
		end
	end

	return false;
end

-- ==========================================================

-- Station Instance Accessors

-- ==========================================================

function GetTradeYieldDeltaPerTurn(stationID, yieldType)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.GetTradeYieldDeltaPerTurn ~= nil) then
		return stationTable.GetTradeYieldDeltaPerTurn(station, yieldType);
	else
		return 0;
	end
end

function GetStationConquestYieldReward(stationID, yieldType)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.GetStationConquestYieldReward ~= nil) then
		return stationTable.GetStationConquestYieldReward(station, yieldType);
	else
		return 0;
	end
end

function GetStationMaxLevel(stationID)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.GetMaxLevel ~= nil) then
		return stationTable.GetMaxLevel();
	else
		return DEFAULT_STATION_MAX_LEVEL;
	end
end

-----------------------------------------------------------------------------------
function GetMessageString(stationType, playerType)

	local stationInfo = GameInfo.Stations[stationType];
	local stationData = {
		NameKey = stationInfo.Description,
		Info = stationInfo,
		Level = 1,
	};

	-- If this station already exists, extract its current state data
	local stationInstance = Game.GetStationByType(stationType);
	if (stationInstance ~= nil) then
		stationData.Level = stationInstance:GetLevel();		
	end

	local stationTable = GetStationTable(stationType);
	if (stationTable.GetMessageString ~= nil) then
		return stationTable.GetMessageString(stationData, playerType);
	else
		return "STATION_MESSAGE_UNDEFINED";
	end
end

-----------------------------------------------------------------------------------
function GetLevelString(stationID, playerType)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());

	if (CHECK_DEFAULT_LEVELING_OVERRIDE(station)) then
		return stationTable.GetLevelString(station, playerType);
	else
		return DEFAULT_GET_LEVEL_STRING(station, playerType);
	end
end

-- Get a string describing what the station will offer at its next level
function GetNextLevelString(stationID, playerType)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());

	if (stationTable.GetNextLevelString ~= nil) then
		return stationTable.GetNextLevelString(station, playerType);
	else
		return "";
	end
end

-- Get a string describing what special effects the station has when traded with
function GetTradePopupSummary(stationID, playerType)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.GetTradePopupSummary ~= nil) then
		return stationTable.GetTradePopupSummary(station, playerType);
	else
		return "";
	end
end

function GetShouldStationWithdraw(stationID)
	local station = Game.GetStationByID(stationID);
	local stationTable = GetStationTable(station:GetType());
	if (stationTable.GetShouldStationWithdraw ~= nil) then
		return stationTable.GetShouldStationWithdraw(station);
	else
		-- Default method
		local stationTurnsWithoutTrade = station:GetTurnsWithoutTrade();
		-- Scale withdrawal threshold base on game speed
		local withdrawalThreshold = (BASE_WITHDRAWAL_TURN_THRESHOLD * Game:GetBuildPercent()) / 100;
		return stationTurnsWithoutTrade >= withdrawalThreshold;
	end
end

-- ==========================================================

-- Callbacks

-- ==========================================================

function OnStationCreated(stationID)
	local station = Game.GetStationByID(stationID);
	print("Station Created: " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationCreated ~= nil) then
		return stationTable.OnStationCreated(station);
	end
end

function OnStationDestroyed(stationID)
	local station = Game.GetStationByID(stationID);
	print("Station Destroyed: " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationDestroyed ~= nil) then
		return stationTable.OnStationDestroyed(station);
	end
end

-- Trade Route Created
function OnStationTradeRouteCreated(stationID, tradingCityPlayer, tradingCityID)
	local station = Game.GetStationByID(stationID);
	local tradingCity = Players[tradingCityPlayer]:GetCityByID(tradingCityID);
	print("Trade Route established between " .. tradingCity:GetName() .. " and " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationTradeRouteCreated ~= nil) then
		return stationTable.OnStationTradeRouteCreated(station, tradingCity);
	end
end

-- Trade Route Completed
function OnStationTradeRouteCompleted(stationID, tradingCityPlayer, tradingCityID)
	local station = Game.GetStationByID(stationID);
	local tradingCity = Players[tradingCityPlayer]:GetCityByID(tradingCityID);
	print("Trade Route complete between " .. tradingCity:GetName() .. " and " .. station:GetName());

	print("Granting Diplomatic Capital.")
	Players[tradingCityPlayer]:ChangeDiplomaticCapital(CapitalTradeRouteComplete)

	-- Notification
	if Players[tradingCityPlayer]:IsHuman() then
		Players[tradingCityPlayer]:AddNotification(NotificationTypes.NOTIFICATION_GENERIC, TextEventTradeComplete, TextEventTradeCompleteHeadline)
	end	

	-- Grant Artifacts
	local artifactType : number = ArtifactUtilities.ChooseArtifactFromCategory(tradingCityPlayer, GameInfo.ArtifactCategories["ARTIFACT_CATEGORY_RYIKA_STATION"].ID)
	
	if Players[tradingCityPlayer]:IsHuman() then
		print("Human Player finished a Trade Route. Trying to grant an Artifact.")
		if(artifactType ~= -1) then
			print("Success!")
			Players[tradingCityPlayer]:AddArtifact(artifactType)
		else
			print("No Artifacts left. :(")
		end
	end


	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationTradeRouteCompleted ~= nil) then
		return stationTable.OnStationTradeRouteCompleted(station, tradingCity);
	end
end

-- Trade Route Circuit Completed (the trade unit has finished one lap to the station and back to its origin city)
function OnStationTradeRouteCircuitComplete(stationID, tradingCityPlayer, tradingCityID)
	local station = Game.GetStationByID(stationID);
	local tradingCity = Players[tradingCityPlayer]:GetCityByID(tradingCityID);
	print("Trade Route circuit complete between " .. tradingCity:GetName() .. " and " .. station:GetName());

	-- Chance to grant an Artifact
	if Players[tradingCityPlayer]:IsHuman() then	
		local artifactType2 : number = ArtifactUtilities.ChooseArtifactFromCategory(tradingCityPlayer, GameInfo.ArtifactCategories["ARTIFACT_CATEGORY_RYIKA_STATION"].ID)
		local ArtifactRoll = Map.Rand(100, "Artifact Roll");
		print("Trade Route Owner was human, rolling for Artifact: " .. ArtifactRoll .. " - must be <= " .. CircuitChance .. "!")
		if ArtifactRoll <= CircuitChance then
			if(artifactType ~= -1) then
				print("Success!")
				Players[tradingCityPlayer]:AddArtifact(artifactType2)
			else
				print("Success! ...but no Artifacts left. :(")
			end
		else
			print("Failed.")
		end
	end 

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationTradeRouteCircuitComplete ~= nil) then
		return stationTable.OnStationTradeRouteCircuitComplete(station, tradingCity);
	end
end

-- Trade Route Cancelled (the station or the city was detroyed or conquered)
function OnStationTradeRouteCancelled(stationID, tradingCityPlayer, tradingCityID)
	local station = Game.GetStationByID(stationID);
	local tradingCity = Players[tradingCityPlayer]:GetCityByID(tradingCityID);
	print("Trade Route cancelled between " .. tradingCity:GetName() .. " and " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationTradeRouteCancelled ~= nil) then
		return stationTable.OnStationTradeRouteCancelled(station, tradingCity);
	end
end

-- Trade Route Destroyed (the trade unit was plundered)
function OnStationTradeRouteDestroyed(stationID, tradingCityPlayer, tradingCityID)
	local station = Game.GetStationByID(stationID);
	local tradingCity = Players[tradingCityPlayer]:GetCityByID(tradingCityID);
	print("Trade Route destroyed between " .. tradingCity:GetName() .. " and " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationTradeRouteDestroyed ~= nil) then
		return stationTable.OnStationTradeRouteDestroyed(station, tradingCity);
	end
end

-- Do Station Turn ( called from CvStation::DoTurn() )
function DoStationTurn(stationID)
	local station = Game.GetStationByID(stationID);
	print("Station Turn: " .. station:GetName());

	local stationTable = GetStationTable(station:GetType());
	if (stationTable.OnStationDoTurn ~= nil) then
		return stationTable.OnStationDoTurn(station);
	end
end

-- Update Station Level ( called from CvStation::DoTurn() )
function UpdateStationLevel(stationID)
	local station = Game.GetStationByID(stationID);
	print("Updating Station Level: " .. station:GetName());
	local stationTable = GetStationTable(station:GetType());

	if (CHECK_DEFAULT_LEVELING_OVERRIDE(station)) then
		return stationTable.UpdateStationLevel(station);
	else
		return DEFAULT_UPDATE_LEVEL(station);
	end
end

-- ==========================================================

-- Utility functions

-- ==========================================================
function GetStationTable(stationType)
	return g_Stations[stationType];
end

-- ==========================================================
-- Load station scripts
-- ==========================================================
do
	-- Stations
	for stationInfo in GameInfo.Stations() do

		local stationTable;
		local path;

		for addin in Modding.GetActivatedModEntryPoints("MinorPowersAddin") do
			if string.find(addin.File, stationInfo.ScriptName) then
				local addinFile = Modding.GetEvaluatedFilePath(addin.ModID, addin.Version, addin.File);
				path = addinFile.EvaluatedPath;
				print("Loaded MinorPowers script for ''"..Locale.ConvertTextKey(stationInfo.Description).."''")
				print("Added modded MinorPowers script ''" ..addin.Name.."'' with ModID "..addin.ModID)
				print("addinpath: "..path)
				break
			end
		end

		
		if not path then
			path = "assets\\Gameplay\\Lua\\MinorPowers\\" .. stationInfo.ScriptName .. ".lua";
			print("Loaded MinorPowers script for ''"..Locale.ConvertTextKey(stationInfo.Description).."''")
			print("Script loaded from game's Assets folder")
		end

		if path then
			stationTable = dofile(path);
		end
		if (stationTable ~= nil) then
			-- Add some additional data to the station table
			stationTable.Info = stationInfo;

			-- Insert the station into our global list
			g_Stations[stationInfo.ID] = stationTable;
		else
			print("Error: Unable to load the script for Minor Power ''"..Locale.ConvertTextKey(stationInfo.Description))
		end
	end
end